#!/usr/bin/env python
#-*-coding: utf-8 -*-
# Copyright (c) 2019 Synology Inc. All rights reserved.

import sys, enum, errno, os, stat, signal, json, socket, threading
from hypervisor import *

# copied from include/vmbk/defines.h
HYPERVISOR_SERVER_SOCKET = '/run/synoabk/hypervisor_server.sock'
VM_SYNOABK_TEMP = '/tmp/synoabk/vm'

# copied from ActiveBackup-core/lib/record/config_db.sql
class HostType(enum.Enum):
    NONE         = 0
    ESXi         = 1
    vCenter      = 2
    HyperV       = 3
    SystemCenter = 4

import logging
import logging.handlers
logger = logging.getLogger(defines.VM_LOGGER_NAME)
logger.setLevel(logging.DEBUG)

handler = logging.handlers.SysLogHandler(address='/dev/log')
handler.setLevel(logging.INFO)

formatter = logging.Formatter('[%(levelname)s] synoabk_hypervisor_server::%(filename)s:%(lineno)d(%(process)d,%(thread)d) %(message)s')
handler.setFormatter(formatter)

logger.addHandler(handler)

# run daemon
if __name__ == '__main__':
    reload(sys)
    sys.setdefaultencoding('utf-8')

    serverLock = threading.Lock()
    serverPool = {}

    def GetServer(inventory):
        serverLock.acquire()
        try:
            inventoryId = inventory['inventory_id']
            hosttype    = inventory['host_type']
            hostaddr    = inventory['host_addr']
            username    = inventory['auth_user']
            password    = inventory['auth_password']
            thumbprint  = inventory['auth_thumbprint']
            port        = inventory['port_webapi']
            protocol    = inventory['protocol']

            if inventoryId != 0:
                oldInventory, server, taskPool = serverPool.get(inventoryId) or (None, None, {})
            else:
                oldInventory, server, taskPool = None, None, {}

            if  hosttype == HostType.ESXi.value or \
                hosttype == HostType.vCenter.value:
                if  server == None or \
                    (oldInventory['host_addr'] != hostaddr or \
                    oldInventory['auth_user'] != username or \
                    oldInventory['auth_password'] != password or \
                    oldInventory['auth_thumbprint'] != thumbprint or \
                    oldInventory['protocol'] != protocol or \
                    oldInventory['port_webapi'] != port):
                    server = vsphere_server.Main(hostaddr, username, password, thumbprint, port, protocol, taskPool)
            elif hosttype == HostType.HyperV.value:
                server = hyperv_server.Main(hostaddr, username, password, thumbprint, port, protocol, taskPool)
            else:
                raise Exception('Unknown Host Type %d' % hosttype)

            serverPool[inventoryId] = (inventory, server, taskPool)
        finally:
            serverLock.release()

        return server

    def HandleRequest(request):
        command = request['command']
        argument = request['argument']
        inventory = request['inventory']
        server = GetServer(inventory)
        logger.info(command)
        logger.info({k:v for k,v in argument.iteritems() if k not in ['vm_user', 'vm_pwd']} if argument else None)
        return server.dispatch(command, argument)

    def HandleConnection(connection, address):
        try:
            while True:
                data = connection.recv(4)
                if len(data) != 4: break

                n = sum([ord(data[i])*(256**i) for i in range(4)])

                data = connection.recv(n)
                if len(data) != n: break

                data = json.dumps(HandleRequest(json.loads(data.decode('utf-8', 'ignore'))))
                n = len(data)
                data = ''.join([chr(n/(256**i)%256) for i in range(4)]) + data

                connection.sendall(data)
        finally:
            connection.close()

    def HandleSignal(sig, frame):
        if sig == signal.SIGPIPE:
            return
        serverPool.clear()
        os._exit(0)

    print('Hypervisor Server Daemon Start! ^.^')

    try:
        os.unlink(HYPERVISOR_SERVER_SOCKET)
    except Exception:
        pass

    try:
        os.makedirs(os.path.dirname(HYPERVISOR_SERVER_SOCKET))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.bind(HYPERVISOR_SERVER_SOCKET)
    sock.listen(16)

    os.chown(HYPERVISOR_SERVER_SOCKET, int(os.environ['PKG_UID']), int(os.environ['PKG_GID']))
    os.chmod(HYPERVISOR_SERVER_SOCKET, stat.S_IRUSR + stat.S_IWUSR)

    try:
        os.mkdir(VM_SYNOABK_TEMP)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    os.chown(VM_SYNOABK_TEMP, int(os.environ['PKG_UID']), int(os.environ['PKG_GID']))
    os.chmod(VM_SYNOABK_TEMP, stat.S_IRWXU + stat.S_IRGRP + stat.S_IXGRP + stat.S_IROTH + stat.S_IXOTH)

    for sig in (signal.SIGINT, signal.SIGTERM, signal.SIGQUIT, signal.SIGPIPE):
        signal.signal(sig, HandleSignal)

    while True:
        threading.Thread(target=HandleConnection, args=sock.accept()).start()
